"""Evaluation metric functions for Physionet Challenge.

The functions here have been copied from:
physionetchallenges/evaluation-2021/evaluate_model.py
"""

import numpy as np

SINUS_RHYTHM = frozenset(['426783006'])


def compute_accuracy(labels, outputs):
  """Compute recording-wise accuracy."""
  num_recordings, num_classes = np.shape(labels)  # pylint:disable=unused-variable

  num_correct_recordings = 0
  for i in range(num_recordings):
    if np.all(labels[i, :] == outputs[i, :]):
      num_correct_recordings += 1

  return float(num_correct_recordings) / float(num_recordings)


def compute_confusion_matrices(labels, outputs, normalize=False):  # pylint:disable=g-doc-args
  """Compute confusion matrices.

  Compute a binary confusion matrix for each class k:

      [TN_k FN_k]
      [FP_k TP_k]

  If the normalize variable is set to true, then normalize the contributions
  to the confusion matrix by the number of labels per recording.

  Returns:
  """
  num_recordings, num_classes = np.shape(labels)

  if not normalize:
    A = np.zeros((num_classes, 2, 2))  # pylint: disable=invalid-name
    for i in range(num_recordings):
      for j in range(num_classes):
        if labels[i, j] == 1 and outputs[i, j] == 1:  # TP
          A[j, 1, 1] += 1
        elif labels[i, j] == 0 and outputs[i, j] == 1:  # FP
          A[j, 1, 0] += 1
        elif labels[i, j] == 1 and outputs[i, j] == 0:  # FN
          A[j, 0, 1] += 1
        elif labels[i, j] == 0 and outputs[i, j] == 0:  # TN
          A[j, 0, 0] += 1
        else:  # This condition should not happen.
          raise ValueError('Error in computing the confusion matrix.')
  else:
    A = np.zeros((num_classes, 2, 2))  # pylint: disable=invalid-name
    for i in range(num_recordings):
      normalization = float(max(np.sum(labels[i, :]), 1))
      for j in range(num_classes):
        if labels[i, j] == 1 and outputs[i, j] == 1:  # TP
          A[j, 1, 1] += 1.0 / normalization
        elif labels[i, j] == 0 and outputs[i, j] == 1:  # FP
          A[j, 1, 0] += 1.0 / normalization
        elif labels[i, j] == 1 and outputs[i, j] == 0:  # FN
          A[j, 0, 1] += 1.0 / normalization
        elif labels[i, j] == 0 and outputs[i, j] == 0:  # TN
          A[j, 0, 0] += 1.0 / normalization
        else:  # This condition should not happen.
          raise ValueError('Error in computing the confusion matrix.')

  return A


def compute_f_measure(labels, outputs):
  """Compute macro F-measure."""
  num_recordings, num_classes = np.shape(labels)  # pylint:disable=unused-variable

  A = compute_confusion_matrices(labels, outputs)  # pylint:disable=invalid-name

  f_measure = np.zeros(num_classes)
  for k in range(num_classes):
    tp, fp, fn, tn = A[k, 1, 1], A[k, 1, 0], A[k, 0, 1], A[k, 0, 0]  # pylint:disable=unused-variable
    if 2 * tp + fp + fn:
      f_measure[k] = float(2 * tp) / float(2 * tp + fp + fn)
    else:
      f_measure[k] = float('nan')

  if np.any(np.isfinite(f_measure)):
    macro_f_measure = np.nanmean(f_measure)
  else:
    macro_f_measure = float('nan')

  return macro_f_measure, f_measure


def compute_auc(labels, outputs):
  """Compute macro AUROC and macro AUPRC."""
  num_recordings, num_classes = np.shape(labels)

  # Compute and summarize the confusion matrices for each class across at
  # distinct output values.
  auroc = np.zeros(num_classes)
  auprc = np.zeros(num_classes)

  for k in range(num_classes):
    # We only need to compute TPs, FPs, FNs, and TNs at distinct output values.
    thresholds = np.unique(outputs[:, k])
    thresholds = np.append(thresholds, thresholds[-1] + 1)
    thresholds = thresholds[::-1]
    num_thresholds = len(thresholds)

    # Initialize the TPs, FPs, FNs, and TNs.
    tp = np.zeros(num_thresholds)
    fp = np.zeros(num_thresholds)
    fn = np.zeros(num_thresholds)
    tn = np.zeros(num_thresholds)
    fn[0] = np.sum(labels[:, k] == 1)
    tn[0] = np.sum(labels[:, k] == 0)

    # Find the indices that result in sorted output values.
    idx = np.argsort(outputs[:, k])[::-1]

    # Compute the TPs, FPs, FNs, and TNs for class k across thresholds.
    i = 0
    for j in range(1, num_thresholds):
      # Initialize TPs, FPs, FNs, and TNs using values at previous threshold.
      tp[j] = tp[j - 1]
      fp[j] = fp[j - 1]
      fn[j] = fn[j - 1]
      tn[j] = tn[j - 1]

      # Update the TPs, FPs, FNs, and TNs at i-th output value.
      while i < num_recordings and outputs[idx[i], k] >= thresholds[j]:
        if labels[idx[i], k]:
          tp[j] += 1
          fn[j] -= 1
        else:
          fp[j] += 1
          tn[j] -= 1
        i += 1

    # Summarize the TPs, FPs, FNs, and TNs for class k.
    tpr = np.zeros(num_thresholds)
    tnr = np.zeros(num_thresholds)
    ppv = np.zeros(num_thresholds)
    for j in range(num_thresholds):
      if tp[j] + fn[j]:
        tpr[j] = float(tp[j]) / float(tp[j] + fn[j])
      else:
        tpr[j] = float('nan')
      if fp[j] + tn[j]:
        tnr[j] = float(tn[j]) / float(fp[j] + tn[j])
      else:
        tnr[j] = float('nan')
      if tp[j] + fp[j]:
        ppv[j] = float(tp[j]) / float(tp[j] + fp[j])
      else:
        ppv[j] = float('nan')

    # Compute AUROC as the area under a piecewise linear function with TPR/
    # sensitivity (x-axis) and TNR/specificity (y-axis) and AUPRC as the area
    # under a piecewise constant with TPR/recall (x-axis) and PPV/precision
    # (y-axis) for class k.
    for j in range(num_thresholds - 1):
      auroc[k] += 0.5 * (tpr[j + 1] - tpr[j]) * (tnr[j + 1] + tnr[j])
      auprc[k] += (tpr[j + 1] - tpr[j]) * ppv[j + 1]

  # Compute macro AUROC and macro AUPRC across classes.
  if np.any(np.isfinite(auroc)):
    macro_auroc = np.nanmean(auroc)
  else:
    macro_auroc = float('nan')
  if np.any(np.isfinite(auprc)):
    macro_auprc = np.nanmean(auprc)
  else:
    macro_auprc = float('nan')

  return macro_auroc, macro_auprc, auroc, auprc


def compute_modified_confusion_matrix(labels, outputs):
  """Compute a modified confusion matrix for multi-class, multi-label tasks."""
  # Compute a binary multi-class, multi-label confusion matrix, where the rows
  # are the labels and the columns are the outputs.
  num_recordings, num_classes = np.shape(labels)
  A = np.zeros((num_classes, num_classes))  # pylint: disable=invalid-name

  # Iterate over all of the recordings.
  for i in range(num_recordings):
    # Calculate the number of positive labels and/or outputs.
    normalization = float(
        max(np.sum(np.any((labels[i, :], outputs[i, :]), axis=0)), 1))
    # Iterate over all of the classes.
    for j in range(num_classes):
      # Assign full and/or partial credit for each positive class.
      if labels[i, j]:
        for k in range(num_classes):
          if outputs[i, k]:
            A[j, k] += 1.0 / normalization

  return A


def compute_challenge_metric(weights, labels, outputs, classes, sinus_rhythm):
  """Compute the evaluation metric for the Challenge."""
  num_recordings, num_classes = np.shape(labels)
  if sinus_rhythm in classes:
    sinus_rhythm_index = classes.index(sinus_rhythm)
  else:
    raise ValueError('The sinus rhythm class is not available.')

  # Compute the observed score.
  A = compute_modified_confusion_matrix(labels, outputs)  # pylint:disable=invalid-name
  observed_score = np.nansum(weights * A)

  # Compute the score for the model that always chooses the correct label(s).
  correct_outputs = labels
  A = compute_modified_confusion_matrix(labels, correct_outputs)  # pylint:disable=invalid-name
  correct_score = np.nansum(weights * A)

  # Compute the score for the model that always chooses the sinus rhythm class.
  inactive_outputs = np.zeros((num_recordings, num_classes), dtype=np.bool)
  inactive_outputs[:, sinus_rhythm_index] = 1
  A = compute_modified_confusion_matrix(labels, inactive_outputs)  # pylint:disable=invalid-name
  inactive_score = np.nansum(weights * A)

  if correct_score != inactive_score:
    normalized_score = float(observed_score -
                             inactive_score) / float(correct_score -
                                                     inactive_score)
  else:
    normalized_score = 0.0

  return normalized_score


def compute_cinc_score(y_true, y_pred, classes, weights):
  """Wrapper function to calculate CinC score.

  To get "classes" and "weights" that are used as input here, you should use the
  load_weights() function in load_weights.py (for google3) or
  load_weights_gcp.py (for GCP).

  Args:
    y_true: Dictionary of arrays of size num_sample, where each array is the
      true binary output keyed by its label (class).
    y_pred: Dictionary of arrays of size num_sample, where each array is the
      model prediction in binary format, keyed by its label (class).
    classes: Hierarchical list of all classes, where each element is a set of
      labels that fall into the same category when computing the performance.
    weights: Confusion matrix weighting array.

  Returns:
    Evaluation score.
  """
  n_sample = y_pred[list(y_pred.keys())[0]].shape[0]
  binary_outputs = np.zeros((n_sample, len(classes)))
  true_labels = np.zeros((n_sample, len(classes)))
  for i, class_set in enumerate(classes):
    binary_outputs[:, i] = np.logical_or.reduce(
        [y_pred[c] for c in class_set])
    true_labels[:, i] = np.logical_or.reduce([y_true[c] for c in class_set])
  # Calculate the challenge metric.
  challenge_metric = compute_challenge_metric(
      weights, true_labels, binary_outputs, classes, SINUS_RHYTHM)
  return challenge_metric
