"""Autoencoder-based classifier for 12-lead ECG data.
"""
import time
import numpy as np

import tensorflow as tf
import tensorflow.keras.layers as kl

import metrics
import preprocess

# Additional training and validation parameters during training
TRAIN_RANDOM_SEED = 866
VALIDATION_STEPS = 1
VALIDATION_FREQ = 1
EARLY_STOPPING_PATIENCE_AUTOENCODER = 15
EARLY_STOPPING_PATIENCE_CLASSIFIER = 15


def clone_encoder(encoder, input_shape):
  """Create a true deep copy of input encoder for transfer learning."""
  assert len(input_shape) == 2
  new_encoder = Encoder(embedding_dim=encoder.embedding_dim)
  new_encoder.build((None, input_shape[0], input_shape[1]))
  new_encoder.set_weights(encoder.get_weights())
  return new_encoder


class PreProcessECG(tf.keras.layers.Layer):
  """Preprocess ECG leads."""

  def __init__(self, lead_names=preprocess.LEADS):
    super().__init__()
    self.lead_names = lead_names

  def call(self, inputs):
    input_copy = {k: inputs[k]
                  for k in inputs if k not in self.lead_names}
    ecg = []
    for lead in self.lead_names:
      ecg.append(inputs[lead])
    input_copy['ecg'] = tf.stack(ecg, axis=2)
    return input_copy


class PreProcessContext(tf.keras.layers.Layer):
  """Preprocess data for inference."""

  def call(self, inputs):
    input_copy = {k: inputs[k] for k in inputs}
    input_copy['age'] = (tf.cast(input_copy['age'], tf.float32) - 50.) / 25.
    input_copy['female'] = tf.cast(
        tf.equal(input_copy['sex'], tf.constant('Female')), tf.float32)
    return input_copy


class Encoder(tf.keras.Model):
  """Class representing an encoder for an autoencoder model."""

  def __init__(self, embedding_dim=16, encoder_layers=None, name='encoder'):
    super(Encoder, self).__init__(name=name)
    self.embedding_dim = embedding_dim
    if encoder_layers is None:
      encoder_layers = [
          kl.Conv1D(64, 40, strides=1, padding='same', activation=tf.nn.swish),
          kl.MaxPool1D(pool_size=2),
          kl.Conv1D(32, 40, strides=1, padding='same', activation=tf.nn.swish),
          kl.MaxPool1D(pool_size=2),
          kl.Conv1D(24, 40, strides=1, padding='same', activation=tf.nn.swish),
          kl.MaxPool1D(pool_size=2),
          kl.Conv1D(16, 20, strides=1, padding='same', activation=tf.nn.swish),
          kl.MaxPool1D(pool_size=2),
          kl.Flatten(data_format='channels_last'),
          kl.Dense(max(4 * embedding_dim, 512)),
          kl.Dense(embedding_dim),
      ]
    self.encoder_layers = encoder_layers
    self.net = tf.keras.Sequential(encoder_layers)

  def get_config(self):
    return {
        'embedding_dim': self.embedding_dim,
        'name': self.name,
        'encoder_layers': None,
    }

  @classmethod
  def from_config(cls, config):
    return cls(**config)

  @tf.function()
  def call(self, inputs):
    return self.net(inputs)


class Decoder(tf.keras.Model):
  """Class representing a decoder for an autoencoder model."""

  def __init__(self, output_shape, layers=None, name='decoder'):
    super(Decoder, self).__init__(name=name)
    self.decoder_output_shape = output_shape
    if layers is None:
      layers = [
          kl.Dense(512),
          kl.Dense(output_shape[0] // 16 * 16),
          kl.Reshape((output_shape[0] // 16, 16)),
          kl.UpSampling1D(2),
          kl.Conv1D(24, 20, strides=1, padding='same', activation=tf.nn.swish),
          kl.UpSampling1D(2),
          kl.Conv1D(32, 40, strides=1, padding='same', activation=tf.nn.swish),
          kl.UpSampling1D(2),
          kl.Conv1D(64, 40, strides=1, padding='same', activation=tf.nn.swish),
          kl.UpSampling1D(2),
          kl.Conv1D(output_shape[-1], 100, strides=1, padding='same',
                    activation=None),
      ]
    self.net = tf.keras.Sequential(layers)

  @tf.function()
  def call(self, inputs):
    return self.net(inputs)


class AutoEncoder(tf.keras.Model):
  """Keras model for an autoencoder using 12-lead ECG signals as input.

  Attributes:
    encoder: Encoder model.
    decoder: Decoder model.
    name: Model name.
  """

  def __init__(self, encoder, decoder, preprocess_ecg_in_model=False,
               use_custom_loss=False, name='autoencoder'):
    super(AutoEncoder, self).__init__(name=name)
    self.preprocess_ecg_in_model = preprocess_ecg_in_model
    self.pp_ecg = PreProcessECG()
    self.use_custom_loss = use_custom_loss
    self.encoder = encoder
    self.decoder = decoder

  def call(self, inputs, training=False):
    if self.preprocess_ecg_in_model:
      inputs_copy = self.pp_ecg(inputs)
    else:
      inputs_copy = {k: inputs[k] for k in inputs}
    y_pred = self.decoder(self.encoder(inputs_copy['ecg']))
    if self.use_custom_loss:
      self.add_loss(
          tf.reduce_mean(tf.square(y_pred - inputs_copy['ecg']), axis=[1, 2]))
      self.add_metric(
          tf.reduce_mean(tf.abs(y_pred - inputs_copy['ecg']), axis=[1, 2]),
          name='mae', aggregation='mean')
      self.add_metric(
          tf.reduce_mean(tf.square(y_pred - inputs_copy['ecg']), axis=[1, 2]),
          name='mse', aggregation='mean')
    return y_pred


class Classifier(tf.keras.Model):
  """Keras model for a multi-output classifier using the low-dimensiontal representation of the ECG signal provided by the autoencoder.

  Attributes:
    encoder: Encoder model.
    labels: List of labels corresponding to different outputs.
    input_shape: Shape of the input to the model.
    regularization_kwargs: Regularization kwargs.
    output_bias: List of bias values for each output. This should have the same
      length as labels.
  """

  def __init__(self, encoder, labels, input_shape, regularization_kwargs=None,
               output_bias=None, preprocess_ecg_in_model=False,
               use_age=False, use_sex=False):
    super(Classifier, self).__init__()
    self.preprocess_ecg_in_model = preprocess_ecg_in_model
    self.encoder = clone_encoder(encoder, input_shape)
    self.pp_ecg = PreProcessECG()
    self.pp_context = PreProcessContext()
    self.labels = labels
    self.use_age = use_age
    self.use_sex = use_sex
    if not regularization_kwargs:
      regularization_kwargs = {}
    layers = [
        kl.Dense(512, activation=tf.nn.swish, **regularization_kwargs),
        kl.Dropout(0.2),
        kl.Dense(128, activation=tf.nn.swish, **regularization_kwargs),
        kl.Dropout(0.2),
        kl.Dense(64, activation=tf.nn.swish, **regularization_kwargs),
    ]
    self.classifier_trunk = tf.keras.Sequential(layers)
    self.leaf_layer = []
    if output_bias is not None:
      assert len(output_bias) == len(labels)
    else:
      output_bias = {str(int(l)): 0 for l in labels}
    for label in labels:
      self.leaf_layer.append(kl.Dense(
          1, activation='sigmoid', name=str(int(label)),
          bias_initializer=tf.keras.initializers.Constant(
              output_bias[str(int(label))])))

  def call(self, inputs, training=False):
    if self.preprocess_ecg_in_model:
      inputs_copy = self.pp_ecg(inputs)
    else:
      inputs_copy = {k: inputs[k] for k in inputs}
    inputs_copy = self.pp_context(inputs_copy)
    embedded = self.encoder(inputs_copy['ecg'])
    last_input = self.classifier_trunk(embedded, training=training)
    if self.use_age:
      last_input = kl.Concatenate(axis=1)([last_input, inputs_copy['age']])
    if self.use_sex:
      last_input = kl.Concatenate(axis=1)([last_input, inputs_copy['female']])
    return {l.name: l(last_input) for l in self.leaf_layer}

  def encoders_trainable(self, trainable: bool):
    self.encoder.trainable = trainable


def optimize_threshold_cinc(y_true, y_pred, classes, weights, thresholds):
  """Finds the optimal threshold for prediction probabilities using the CinC metric."""
  score = []
  for th in thresholds:
    y_pred_binary = {
        key: (value > th).astype(int) for key, value in y_pred.items()}
    y_true_agg, y_pred_agg = preprocess.aggregate_records(y_true, y_pred_binary)
    score.append(
        metrics.compute_cinc_score(y_true_agg, y_pred_agg, classes, weights))
  threshold_opt = thresholds[np.argmax(score)]
  return threshold_opt


def optimize_threshold_ind(y_true, y_pred, classes, thresholds):
  """Finds the optimal thresholds for prediction probabilities for each label independently."""
  threshold_opt = {}
  for c in preprocess.flatten_classes(classes):
    score = []
    yt = y_true[c]
    for th in thresholds:
      yp = (y_pred[c] > th).astype(int)
      tp = np.sum((yt == 1) & (yp == 1))
      tn = np.sum((yt == 0) & (yp == 0))
      fp = np.sum((yt == 0) & (yp == 1))
      fn = np.sum((yt == 1) & (yp == 0))
      if (fp + fn) == 0 or (tp + fn) == 0:
        score.append(np.nan)
      else:
        fpr = fp / (fp + tn)
        tpr = tp / (tp + fn)
        score.append(np.sqrt((1 - fpr) ** 2 + tpr ** 2))
    score = [s for s in score if np.isfinite(s)]
    if score:
      threshold_opt[c] = thresholds[np.argmax(score)]
    else:
      threshold_opt[c] = 0.5
  return threshold_opt


def optimize_threshold(
    y_true, y_pred, classes, weights, threshold_min=0.01, threshold_max=0.35,
    n=100, threshold_optimization='cinc'):
  """Finds the optimal threshold for prediction probabilities."""
  thresholds = np.linspace(threshold_min, threshold_max, n)
  if threshold_optimization == 'independent':
    threshold_opt = optimize_threshold_ind(y_true, y_pred, classes, thresholds)
  elif threshold_optimization == 'cinc':
    threshold_opt = optimize_threshold_cinc(
        y_true, y_pred, classes, weights, thresholds)
  else:
    raise ValueError(
        f'Unknown threshold optimization method "{threshold_optimization}"')
  return threshold_opt


def fit_autoencoder(data_train, data_val, epochs, input_shape, embedding_dim,
                    lr, xmanager_callback, batch_size=None,
                    preprocess_ecg_in_model=False, use_custom_loss=False,
                    debug_mode=False):
  """Trains the autoencoder part of the model."""
  if not use_custom_loss:
    assert not preprocess_ecg_in_model, (
        'ecg preprocess must be False to use keras loss')
  encoder = Encoder(embedding_dim=embedding_dim)
  decoder = Decoder(output_shape=input_shape)
  model = AutoEncoder(encoder, decoder,
                      preprocess_ecg_in_model=preprocess_ecg_in_model,
                      use_custom_loss=use_custom_loss)
  optimizer = tf.keras.optimizers.Adam(learning_rate=lr)
  if not use_custom_loss:
    model.compile(optimizer=optimizer, loss='mse', metrics=['mae'])
  else:
    model.compile(optimizer=optimizer)
  # Add early stopping callback if validation data is given.
  callbacks = []
  validation_steps = None
  validation_freq = None
  if data_val is not None:
    callbacks.append(tf.keras.callbacks.EarlyStopping(
        'val_loss', min_delta=0, restore_best_weights=True,
        patience=EARLY_STOPPING_PATIENCE_AUTOENCODER))
    validation_steps = VALIDATION_STEPS
    validation_freq = VALIDATION_FREQ
  if xmanager_callback:
    callbacks.append(xmanager_callback)
  if isinstance(data_train, tuple):  # Cloud implementation
    x_train, y_train = data_train
    history = model.fit(
        x_train, y_train, epochs=epochs, verbose=debug_mode,
        validation_data=data_val, callbacks=callbacks,
        validation_steps=validation_steps, validation_freq=validation_freq,
        batch_size=batch_size)
  else:  # XManager implementation
    if not preprocess_ecg_in_model:
      data_train = data_train.map(lambda x: (x, x['ecg']))
      data_val = data_val.map(lambda x: (x, x['ecg']))
    history = model.fit(
        data_train, epochs=epochs, verbose=debug_mode, validation_data=data_val,
        callbacks=callbacks, validation_steps=validation_steps,
        validation_freq=validation_freq)

  return history, encoder, model


# TODO(b/178851571): use cinc score as validation metric (and maybe stopping
# criterion)
def fit_classifier(
    encoder, data_train, data_val, classes, input_shape, lr, epochs,
    epochs_extra, pos_neg_ratio, train_encoder=True, xmanager_callbacks=None,
    batch_size=None, preprocess_ecg_in_model=False,
    use_age=False, use_sex=False, imbalance_strategy='bias', debug_mode=False):
  """Fits the classifier."""
  if xmanager_callbacks:
    if train_encoder:
      assert len(xmanager_callbacks) == 2
    else:
      assert len(xmanager_callbacks) == 1
  # Train the classifier.
  classes_flat = preprocess.flatten_classes(classes)
  if imbalance_strategy == 'bias':
    bias = {key: np.log(val) for key, val in pos_neg_ratio.items()}
    loss_weights = None
  elif imbalance_strategy == 'loss':
    bias = None
    # loss = WeightedBinaryCrossEntropy(pos_neg_ratio)
    loss_weights = {key: 1 / val for key, val in pos_neg_ratio.items()}
  else:
    raise ValueError(f'Unknown imbalance strategy "{imbalance_strategy}"')
  model = Classifier(
      encoder, classes_flat, input_shape=input_shape, output_bias=bias,
      preprocess_ecg_in_model=preprocess_ecg_in_model, use_age=use_age,
      use_sex=use_sex)
  model.encoders_trainable(False)
  optimizer = tf.keras.optimizers.Adam(learning_rate=lr)
  model.compile(
      optimizer=optimizer, loss='binary_crossentropy', metrics=['AUC'],
      loss_weights=loss_weights)
  # Add early stopping callback if validation data is given.
  callbacks = []
  validation_steps = None
  validation_freq = None
  if data_val is not None:
    callbacks.append(tf.keras.callbacks.EarlyStopping(
        'val_loss', min_delta=0, restore_best_weights=True,
        patience=EARLY_STOPPING_PATIENCE_CLASSIFIER, mode='min'))
    validation_steps = VALIDATION_STEPS
    validation_freq = VALIDATION_FREQ
  if xmanager_callbacks:
    callbacks.append(xmanager_callbacks[0])
  if isinstance(data_train, tuple):  # Cloud implementation
    x_train, y_train = data_train
    history = model.fit(
        x_train, y_train, epochs=epochs, verbose=False,
        validation_data=data_val, callbacks=callbacks,
        validation_steps=validation_steps, validation_freq=validation_freq,
        batch_size=batch_size)
  else:  # XManager implementation
    history = model.fit(
        data_train, epochs=epochs, verbose=False, validation_data=data_val,
        callbacks=callbacks, validation_steps=validation_steps,
        validation_freq=validation_freq)
  history_extra = None
  # Now train a little more with encoder weights unfrozen.
  if train_encoder:
    model.encoders_trainable(True)
    # Encoder retraining LR set to 1/10 of classifier's LR.
    optimizer = tf.keras.optimizers.Adam(learning_rate=lr / 10)
    model.compile(
        optimizer=optimizer, loss='binary_crossentropy', metrics=['AUC'],
        loss_weights=loss_weights)
    # Add early stopping callback if validation data is given.
    callbacks = []
    if data_val is not None:
      callbacks = [tf.keras.callbacks.EarlyStopping(
          'val_loss', min_delta=0, restore_best_weights=True,
          patience=EARLY_STOPPING_PATIENCE_CLASSIFIER, mode='min')]
    if xmanager_callbacks:
      callbacks.append(xmanager_callbacks[1])
    if isinstance(data_train, tuple):  # Cloud implementation
      x_train, y_train = data_train
      history_extra = model.fit(
          x_train, y_train, epochs=epochs_extra, verbose=debug_mode,
          validation_data=data_val, callbacks=callbacks,
          validation_steps=validation_steps, validation_freq=validation_freq,
          batch_size=batch_size)
    else:  # XManager implementation
      history_extra = model.fit(
          data_train,
          epochs=epochs_extra, verbose=False,
          validation_data=data_val, callbacks=callbacks,
          validation_steps=validation_steps, validation_freq=validation_freq)
    model.encoders_trainable(False)
  return history, history_extra, model


def fit_model(
    dataset_train, dataset_val, classes, weights, xmanager_callbacks,
    pos_neg_ratio, input_shape, embedding_dim, encoder_lr, classifier_lr,
    encoder_epochs, classifier_epochs, classifier_extra_epochs,
    threshold_optimization, preprocess_ecg_in_model=False, use_age=False,
    use_sex=False, use_custom_loss=False, debug_mode=False):
  """Fits the neural network model using training and validation datasets."""
  if xmanager_callbacks is not None:
    assert len(xmanager_callbacks) in [2, 3]
  # Create and train the autoencoder network.
  np.random.seed(TRAIN_RANDOM_SEED)
  tf.random.set_seed(TRAIN_RANDOM_SEED)
  if not preprocess_ecg_in_model:
    dataset_train = dataset_train.map(preprocess.combine_leads)
    dataset_val = dataset_val.map(preprocess.combine_leads)
  hist_ae, encoder, autoencoder = fit_autoencoder(
      data_train=dataset_train,
      data_val=dataset_val,
      epochs=encoder_epochs,
      input_shape=input_shape,
      lr=encoder_lr,
      embedding_dim=embedding_dim,
      xmanager_callback=xmanager_callbacks[0],
      preprocess_ecg_in_model=preprocess_ecg_in_model,
      use_custom_loss=use_custom_loss,
      debug_mode=debug_mode)
  # Create and train the classifier.
  hist_cl, hist_cl_extra, classifier = fit_classifier(
      encoder=encoder,
      data_train=dataset_train.map(
          lambda x: (x, preprocess.parse_labels(x['labels']))),
      data_val=dataset_val.map(
          lambda x: (x, preprocess.parse_labels(x['labels']))),
      classes=classes,
      input_shape=input_shape,
      lr=classifier_lr,
      epochs=classifier_epochs,
      epochs_extra=classifier_extra_epochs,
      pos_neg_ratio=pos_neg_ratio,
      train_encoder=True,
      xmanager_callbacks=xmanager_callbacks[1:],
      preprocess_ecg_in_model=preprocess_ecg_in_model,
      use_age=use_age,
      use_sex=use_sex,
      imbalance_strategy='bias',
      debug_mode=debug_mode)
  # Now find the optimal probability threshold using both train and val sets.
  dataset_trainval = dataset_train.concatenate(dataset_val)
  y_true = preprocess.extract_output_as_dict(dataset_trainval, classes)
  y_pred = classifier.predict(dataset_trainval)
  if threshold_optimization != 'none':
    threshold_opt = optimize_threshold(
        y_true=y_true,
        y_pred=y_pred,
        classes=classes,
        weights=weights,
        threshold_optimization=threshold_optimization)
  else:
    threshold_opt = None
  return classifier, threshold_opt, autoencoder, hist_ae, hist_cl, hist_cl_extra


def fit_model_gcp(
    data, classes, weights, input_shape, embedding_dim, autoencoder_epochs,
    classifier_epochs, classifier_extra_epochs, batch_size, encoder_lr,
    classifier_lr, threshold_optimization, use_age, use_sex, retrain_encoder,
    imbalance_strategy, start_time):
  """Fits the neural network model."""
  # Create and train the autoencoder network.
  np.random.seed(TRAIN_RANDOM_SEED)
  tf.random.set_seed(TRAIN_RANDOM_SEED)
  _, encoder, _ = fit_autoencoder(
      data_train=({'ecg': data['ecg']}, None),
      data_val=None,
      epochs=autoencoder_epochs,
      input_shape=input_shape,
      lr=encoder_lr,
      embedding_dim=embedding_dim,
      xmanager_callback=None,
      batch_size=batch_size,
      use_custom_loss=True)
  print(f'Autoencoder trained: {int(time.perf_counter() - start_time)} sec')
  # Format output and positive/negative ratio for each label.
  y_train_dict = {}
  pos_neg_ratio = {}
  classes_flat = preprocess.flatten_classes(classes)
  for i, lab in enumerate(classes_flat):
    y_train_dict[lab] = data['label'][:, i].astype(int)
    num = data['label'][:, i].sum()
    denom = (data['label'][:, i] == 0).sum()
    if denom and num:
      # pos_neg_ratio[lab] = np.log(num / denom)
      pos_neg_ratio[lab] = num / denom
    else:
      pos_neg_ratio[lab] = 1
  print(f'Output formatted: {int(time.perf_counter() - start_time)} sec')
  # Create and train the classifier.
  _, _, classifier = fit_classifier(
      encoder=encoder,
      data_train=(
          {key: data[key] for key in ['ecg', 'sex', 'age']},
          y_train_dict,
      ),
      data_val=None,
      classes=classes,
      input_shape=input_shape,
      epochs=classifier_epochs,
      epochs_extra=classifier_extra_epochs,
      pos_neg_ratio=pos_neg_ratio,
      train_encoder=retrain_encoder,
      xmanager_callbacks=None,
      batch_size=batch_size,
      lr=classifier_lr,
      use_age=use_age,
      use_sex=use_sex,
      imbalance_strategy=imbalance_strategy)
  print(f'Classifier trained: {int(time.perf_counter() - start_time)} sec')
  y_pred = classifier.predict(data)
  # Add idnum for record aggregation
  y_train_dict['idnum'] = data['idnum']
  # Now find the optimal probability threshold.
  if threshold_optimization != 'none':
    threshold_opt = optimize_threshold(
        y_true=y_train_dict,
        y_pred=y_pred,
        classes=classes,
        weights=weights,
        threshold_optimization=threshold_optimization)
    print(f'Threshold optimized: {int(time.perf_counter() - start_time)} sec')
  else:
    threshold_opt = None
  return classifier, threshold_opt
